#include "application.h"

#include <iostream>

#include "command_line.h"
#include "imgui_impl_sdl3.h"
#include "imgui_internal.h"
#include "filesystem/stb_image.h"
#include "rhi/texture.h"
#include "rhi/vulkan/renderer_vk.h"
#include "spdlog/async.h"
#include "spdlog/spdlog.h"
#include "spdlog/sinks/basic_file_sink.h"

bool g_is_running = true;
bool g_exit_requested = false;
application* g_app_instance = nullptr;

static void glfw_error_callback(int error, const char* description) {
    spdlog::error("Glfw Error {}: {}", error, description);
}

void application::init(const window_params& in_window_params, int argc, char** argv) {
    try {
        async_spdlog_ = spdlog::basic_logger_mt<spdlog::async_factory>("async_file_logger", "logs/log.txt");
    } catch (const spdlog::spdlog_ex& ex) {
        std::cout << "Log init failed: " << ex.what() << std::endl;
    }

    command_line::instance().init(argc, argv);
    init_glfw();
    init_imgui();

    renderer_ = new renderer_vk();

    renderer_->pre_init();

    // new glfw window
    window_ = glfwCreateWindow(in_window_params.width, in_window_params.height, in_window_params.title.c_str(), nullptr,
                               nullptr);
    if (!window_) {
        spdlog::error("Failed to create glfw window");
        return;
    }

    renderer_->init(window_);
    renderer_->resize(in_window_params.width, in_window_params.height);
    g_is_running = true;
}

int application::run() {
    while (!g_exit_requested) {
        glfwPollEvents();
        g_exit_requested = glfwWindowShouldClose(window_);
        if (g_exit_requested)
            break;

        renderer_->new_frame(window_);
        draw_gui();
        renderer_->end_frame(window_);
    }

    shutdown();
    return 0;
}

void application::shutdown() {
    renderer_->shutdown();
    destroy_imgui();
    renderer_->post_shutdown();

    delete renderer_;
    renderer_ = nullptr;

    destroy_glfw();

    g_is_running = false;
}

std::shared_ptr<texture> application::load_texture(const std::string& path) const {
    int width = 0;
    int height = 0;
    unsigned char* image_data = stbi_load(path.c_str(), &width, &height, nullptr, 4);
    if (!image_data) {
        spdlog::error("Failed to load texture: {}", path.c_str());
        return nullptr;
    }
    const auto texture = renderer_->create_texture(image_data, width, height);
    stbi_image_free(image_data);
    return texture;
}

std::shared_ptr<texture>
application::create_texture(const unsigned char* data, const int width, const int height) const {
    return renderer_->create_texture(data, width, height);
}

std::shared_ptr<render_target> application::create_render_target(const int width, const int height,
                                                                 texture_format format) const {
    return renderer_->create_render_target(width, height, format);
}

std::shared_ptr<pixel_shader_drawer> application::create_pixel_shader_drawer() const {
    return renderer_->create_pixel_shader_drawer();
}

void application::init_glfw() {
    glfwSetErrorCallback(glfw_error_callback);
    if (!glfwInit()) {
        spdlog::error("Failed to initialize GLFW");
        return;
    }
}

void application::init_imgui() {
    // Setup Dear ImGui context
    IMGUI_CHECKVERSION();
    init_imgui(ImGui::CreateContext());
    ImGuiIO& io = ImGui::GetIO();
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard; // Enable Keyboard Controls
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableGamepad; // Enable Gamepad Controls
    io.ConfigFlags |= ImGuiConfigFlags_DockingEnable; // Enable Docking
    io.ConfigFlags |= ImGuiConfigFlags_ViewportsEnable; // Enable Multi-Viewport / Platform Windows
    ImGui::StyleColorsDark();
    //ImGui::StyleColorsLight();
}

void application::destroy_glfw() {
    glfwDestroyWindow(window_);
    window_ = nullptr;
    glfwTerminate();
}

void application::destroy_imgui() {
    init_imgui(nullptr);
    ImGui::DestroyContext();
}
